"""Tests for lifecycle hooks."""

from unittest.mock import MagicMock

import pytest
from pydynox import Model, ModelConfig, clear_default_client
from pydynox.attributes import StringAttribute
from pydynox.hooks import (
    HookType,
    after_delete,
    after_load,
    after_save,
    after_update,
    before_delete,
    before_save,
    before_update,
)


@pytest.fixture(autouse=True)
def reset_state():
    """Reset default client before and after each test."""
    clear_default_client()
    yield
    clear_default_client()


@pytest.fixture
def mock_client():
    """Create a mock DynamoDB client."""
    return MagicMock()


def test_hook_decorator_sets_hook_type():
    """Test that decorators set _hook_type attribute."""

    @before_save
    def my_hook(self):
        pass

    assert my_hook._hook_type == HookType.BEFORE_SAVE


@pytest.mark.parametrize(
    "decorator,expected_type",
    [
        pytest.param(before_save, HookType.BEFORE_SAVE, id="before_save"),
        pytest.param(after_save, HookType.AFTER_SAVE, id="after_save"),
        pytest.param(before_delete, HookType.BEFORE_DELETE, id="before_delete"),
        pytest.param(after_delete, HookType.AFTER_DELETE, id="after_delete"),
        pytest.param(before_update, HookType.BEFORE_UPDATE, id="before_update"),
        pytest.param(after_update, HookType.AFTER_UPDATE, id="after_update"),
        pytest.param(after_load, HookType.AFTER_LOAD, id="after_load"),
    ],
)
def test_all_hook_decorators(decorator, expected_type):
    """Test all hook decorators set correct type."""

    @decorator
    def hook(self):
        pass

    assert hook._hook_type == expected_type


def test_model_collects_hooks(mock_client):
    """Test that Model metaclass collects hooks."""

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_save
        def validate(self):
            pass

        @after_save
        def notify(self):
            pass

    assert len(User._hooks[HookType.BEFORE_SAVE]) == 1
    assert len(User._hooks[HookType.AFTER_SAVE]) == 1


def test_model_inherits_hooks(mock_client):
    """Test that hooks are inherited from parent class."""

    class BaseModel(Model):
        model_config = ModelConfig(table="base", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_save
        def base_validate(self):
            pass

    class User(BaseModel):
        model_config = ModelConfig(table="users", client=mock_client)

        @before_save
        def user_validate(self):
            pass

    # Should have both hooks
    assert len(User._hooks[HookType.BEFORE_SAVE]) == 2


def test_before_save_hook_runs(mock_client):
    """Test that before_save hook runs before save."""
    call_order = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_save
        def validate(self):
            call_order.append("before_save")

    User._client_instance = None

    user = User(pk="USER#1")
    user.save()

    assert "before_save" in call_order
    mock_client.put_item.assert_called_once()


def test_after_save_hook_runs(mock_client):
    """Test that after_save hook runs after save."""
    call_order = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @after_save
        def notify(self):
            call_order.append("after_save")

    User._client_instance = None

    user = User(pk="USER#1")
    user.save()

    assert "after_save" in call_order


def test_skip_hooks_on_save(mock_client):
    """Test that skip_hooks=True skips hooks."""
    hook_called = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_save
        def validate(self):
            hook_called.append("called")

    User._client_instance = None

    user = User(pk="USER#1")
    user.save(skip_hooks=True)

    assert len(hook_called) == 0
    mock_client.put_item.assert_called_once()


def test_model_config_skip_hooks_default(mock_client):
    """Test that model_config.skip_hooks=True skips hooks by default."""
    hook_called = []

    class BulkModel(Model):
        model_config = ModelConfig(table="bulk", client=mock_client, skip_hooks=True)
        pk = StringAttribute(hash_key=True)

        @before_save
        def validate(self):
            hook_called.append("called")

    BulkModel._client_instance = None

    item = BulkModel(pk="ITEM#1")
    item.save()

    assert len(hook_called) == 0


def test_model_config_skip_hooks_override(mock_client):
    """Test that skip_hooks=False overrides model_config.skip_hooks=True."""
    hook_called = []

    class BulkModel(Model):
        model_config = ModelConfig(table="bulk", client=mock_client, skip_hooks=True)
        pk = StringAttribute(hash_key=True)

        @before_save
        def validate(self):
            hook_called.append("called")

    BulkModel._client_instance = None

    item = BulkModel(pk="ITEM#1")
    item.save(skip_hooks=False)

    assert len(hook_called) == 1


def test_before_delete_hook_runs(mock_client):
    """Test that before_delete hook runs."""
    hook_called = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_delete
        def check_can_delete(self):
            hook_called.append("before_delete")

    User._client_instance = None

    user = User(pk="USER#1")
    user.delete()

    assert "before_delete" in hook_called


def test_before_update_hook_runs(mock_client):
    """Test that before_update hook runs."""
    hook_called = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

        @before_update
        def validate_update(self):
            hook_called.append("before_update")

    User._client_instance = None

    user = User(pk="USER#1", name="John")
    user.update(name="Jane")

    assert "before_update" in hook_called


def test_hook_can_raise_exception(mock_client):
    """Test that hooks can raise exceptions to stop operation."""

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)
        email = StringAttribute()

        @before_save
        def validate_email(self):
            if not self.email.endswith("@company.com"):
                raise ValueError("Invalid email domain")

    User._client_instance = None

    user = User(pk="USER#1", email="test@gmail.com")

    with pytest.raises(ValueError, match="Invalid email domain"):
        user.save()

    # put_item should not be called
    mock_client.put_item.assert_not_called()


def test_multiple_hooks_run_in_order(mock_client):
    """Test that multiple hooks run in definition order."""
    call_order = []

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)

        @before_save
        def first_hook(self):
            call_order.append("first")

        @before_save
        def second_hook(self):
            call_order.append("second")

    User._client_instance = None

    user = User(pk="USER#1")
    user.save()

    assert call_order == ["first", "second"]



import logging

from pydynox import DynamoDBClient

# Enable INFO level logs for pydynox
logging.basicConfig(level=logging.INFO)

client = DynamoDBClient()

# All operations are logged automatically
client.put_item("users", {"pk": "USER#1", "name": "John"})
# INFO:pydynox:put_item table=users duration_ms=8.2 wcu=1.0

client.get_item("users", {"pk": "USER#1"})
# INFO:pydynox:get_item table=users duration_ms=12.1 rcu=0.5



"""Rate limiting for DynamoDB operations.

Provides token bucket-based rate limiting to prevent throttling.
Two strategies are available:

- FixedRate: Use when you know exactly how much capacity to use
- AdaptiveRate: Auto-adjusts based on throttling feedback

Example:
    >>> from pydynox import DynamoDBClient, FixedRate
    >>> client = DynamoDBClient(rate_limit=FixedRate(rcu=50))
    >>> # All operations now respect the 50 RCU limit

    >>> from pydynox import AdaptiveRate
    >>> client = DynamoDBClient(rate_limit=AdaptiveRate(max_rcu=100))
    >>> # Rate auto-adjusts based on throttling
"""

from __future__ import annotations

from pydynox import pydynox_core

# Re-export Rust classes
FixedRate = pydynox_core.FixedRate
AdaptiveRate = pydynox_core.AdaptiveRate
RateLimitMetrics = pydynox_core.RateLimitMetrics

# Add docstrings for IDE support
FixedRate.__doc__ = """Fixed rate limiter.

Use when you know exactly how much capacity to use.
The rate stays constant unless you change it manually.

Args:
    rcu: Read capacity units per second (optional).
    wcu: Write capacity units per second (optional).
    burst: Burst capacity. Defaults to rate value if not set.

Attributes:
    rcu: The configured RCU rate.
    wcu: The configured WCU rate.
    consumed_rcu: Total RCU consumed so far.
    consumed_wcu: Total WCU consumed so far.
    throttle_count: Number of times throttled.

Example:
    >>> # Read only
    >>> FixedRate(rcu=50)

    >>> # Read and write
    >>> FixedRate(rcu=50, wcu=25)

    >>> # With burst (DynamoDB allows short bursts)
    >>> FixedRate(rcu=50, burst=200)
"""

AdaptiveRate.__doc__ = """Adaptive rate limiter that adjusts based on throttling.

Starts at 50% of max rate. When throttled, reduces by 20%.
When no throttle for 10 seconds, increases by 10%.

Args:
    max_rcu: Maximum read capacity units per second.
    max_wcu: Maximum write capacity units per second (optional).
    min_rcu: Minimum RCU, won't go below this (default: 1).
    min_wcu: Minimum WCU, won't go below this (default: 1).

Attributes:
    current_rcu: Current RCU rate (changes based on throttling).
    current_wcu: Current WCU rate (changes based on throttling).
    max_rcu: Maximum RCU configured.
    max_wcu: Maximum WCU configured.
    consumed_rcu: Total RCU consumed so far.
    consumed_wcu: Total WCU consumed so far.
    throttle_count: Number of times throttled.

Example:
    >>> # Just set the max, it figures out the rest
    >>> AdaptiveRate(max_rcu=100)

    >>> # With write limit too
    >>> AdaptiveRate(max_rcu=100, max_wcu=50)

    >>> # Custom min (won't go below this)
    >>> AdaptiveRate(max_rcu=100, min_rcu=10)
"""

RateLimitMetrics.__doc__ = """Metrics for monitoring rate limiter behavior.

Attributes:
    consumed_rcu: Total RCU consumed.
    consumed_wcu: Total WCU consumed.
    throttle_count: Number of times throttled.

Methods:
    reset(): Reset all metrics to zero.
"""

__all__ = [
    "FixedRate",
    "AdaptiveRate",
    "RateLimitMetrics",
]



"""Batch operations for DynamoDB."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pydynox.client import DynamoDBClient


class BatchWriter:
    """Context manager for batch write operations.

    Collects put and delete operations, then sends them all at once
    when the context exits. Handles splitting into batches of 25 items
    and retrying unprocessed items.

    Example:
        >>> with BatchWriter(client, "users") as batch:
        ...     batch.put({"pk": "USER#1", "sk": "PROFILE", "name": "Alice"})
        ...     batch.put({"pk": "USER#2", "sk": "PROFILE", "name": "Bob"})
        ...     batch.delete({"pk": "USER#3", "sk": "PROFILE"})
    """

    def __init__(self, client: DynamoDBClient, table: str):
        """Create a BatchWriter.

        Args:
            client: The DynamoDBClient to use.
            table: The table name.
        """
        self._client = client
        self._table = table
        self._put_items: list[dict[str, Any]] = []
        self._delete_keys: list[dict[str, Any]] = []

    def __enter__(self) -> BatchWriter:
        """Enter the context manager."""
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any
    ) -> None:
        """Exit the context manager and send all operations."""
        if exc_type is None:
            self.flush()

    def put(self, item: dict[str, Any]) -> None:
        """Add an item to be put.

        Args:
            item: The item to put (as a dict).
        """
        self._put_items.append(item)

    def delete(self, key: dict[str, Any]) -> None:
        """Add a key to be deleted.

        Args:
            key: The key to delete (as a dict with pk and optional sk).
        """
        self._delete_keys.append(key)

    def flush(self) -> None:
        """Send all collected operations to DynamoDB.

        Called automatically when exiting the context manager.
        Can also be called manually to send operations early.
        """
        if not self._put_items and not self._delete_keys:
            return

        self._client.batch_write(
            self._table,
            put_items=self._put_items,
            delete_keys=self._delete_keys,
        )

        # Clear the lists after successful write
        self._put_items = []
        self._delete_keys = []



from pydynox import Model
from pydynox.attributes import StringAttribute


class User(Model):
    class Meta:
        table = "users"

    pk = StringAttribute(hash_key=True)
    sk = StringAttribute(range_key=True)
    name = StringAttribute()
    bio = StringAttribute(null=True)


user = User(pk="USER#123", sk="PROFILE", name="John", bio="Hello world!")

# Check item size
size = user.calculate_size()
print(f"Size: {size.bytes} bytes")
print(f"Size: {size.kb:.2f} KB")
print(f"Percent of limit: {size.percent:.1f}%")
print(f"Over limit: {size.is_over_limit}")



from pydynox import Model, ModelConfig
from pydynox.attributes import StringAttribute
from pydynox.hooks import after_save, before_save


class User(Model):
    model_config = ModelConfig(table="users")

    pk = StringAttribute(hash_key=True)
    email = StringAttribute()
    name = StringAttribute()

    @before_save
    def validate_email(self):
        if not self.email or "@" not in self.email:
            raise ValueError("Invalid email")

    @before_save
    def normalize(self):
        self.email = self.email.lower().strip()
        self.name = self.name.strip()

    @after_save
    def log_save(self):
        print(f"Saved user: {self.pk}")


# Hooks run automatically
user = User(pk="USER#1", email="JOHN@TEST.COM", name="john doe")
user.save()  # Validates, normalizes, then logs

# Skip hooks if needed
user.save(skip_hooks=True)



"""Pydantic integration for pydynox.

Use Pydantic models directly with DynamoDB without redefining your schema.

Example:
    >>> from pydynox.integrations.pydantic import dynamodb_model
    >>> from pydantic import BaseModel
    >>>
    >>> @dynamodb_model(table="users", hash_key="pk", range_key="sk")
    ... class User(BaseModel):
    ...     pk: str
    ...     sk: str
    ...     name: str
    ...     age: int = 0
    >>>
    >>> user = User(pk="USER#1", sk="PROFILE", name="John", age=30)
    >>> user.save()
    >>>
    >>> user = User.get(pk="USER#1", sk="PROFILE")
"""

from __future__ import annotations

from typing import Any, TypeVar

try:
    from pydantic import BaseModel
except ImportError:
    BaseModel = None  # type: ignore

from pydynox.client import DynamoDBClient

T = TypeVar("T")

__all__ = ["dynamodb_model", "from_pydantic"]


def _check_pydantic() -> None:
    """Check if pydantic is installed."""
    if BaseModel is None:
        raise ImportError(
            "pydantic is required for this feature. Install it with: pip install pydynox[pydantic]"
        )


def dynamodb_model(
    table: str,
    hash_key: str,
    range_key: str | None = None,
    region: str | None = None,
    endpoint_url: str | None = None,
) -> Any:
    """Decorator to add DynamoDB operations to a Pydantic model.

    Args:
        table: DynamoDB table name.
        hash_key: Name of the hash key attribute.
        range_key: Name of the range key attribute (optional).
        region: AWS region (optional).
        endpoint_url: Custom endpoint URL for local testing (optional).

    Returns:
        A decorator that adds DynamoDB methods to the class.

    Example:
        >>> @dynamodb_model(table="users", hash_key="pk", range_key="sk")
        ... class User(BaseModel):
        ...     pk: str
        ...     sk: str
        ...     name: str
        ...     age: int = 0
        >>>
        >>> user = User(pk="USER#1", sk="PROFILE", name="John")
        >>> user.save()
        >>>
        >>> user = User.get(pk="USER#1", sk="PROFILE")
    """
    _check_pydantic()

    def decorator(cls: type[T]) -> type[T]:
        if not issubclass(cls, BaseModel):
            raise TypeError(f"{cls.__name__} must be a Pydantic BaseModel subclass")

        # Store metadata
        cls._pydynox_table = table  # type: ignore
        cls._pydynox_hash_key = hash_key  # type: ignore
        cls._pydynox_range_key = range_key  # type: ignore
        cls._pydynox_region = region  # type: ignore
        cls._pydynox_endpoint_url = endpoint_url  # type: ignore
        cls._pydynox_client = None  # type: ignore

        # Add methods
        cls._get_client = classmethod(_get_client_method)  # type: ignore
        cls.get = classmethod(_get_method)  # type: ignore
        cls.save = _save_method  # type: ignore
        cls.delete = _delete_method  # type: ignore
        cls.update = _update_method  # type: ignore
        cls._get_key = _get_key_method  # type: ignore

        return cls

    return decorator


def _get_client_method(cls: type[T]) -> DynamoDBClient:
    """Get or create the DynamoDB client."""
    if cls._pydynox_client is None:  # type: ignore
        cls._pydynox_client = DynamoDBClient(  # type: ignore
            region=cls._pydynox_region,  # type: ignore
            endpoint_url=cls._pydynox_endpoint_url,  # type: ignore
        )
    return cls._pydynox_client  # type: ignore


def _get_method(cls: type[T], **keys: Any) -> T | None:
    """Get an item from DynamoDB by its key.

    Args:
        **keys: The key attributes (hash_key and optional range_key).

    Returns:
        The model instance if found, None otherwise.
    """
    client = cls._get_client()  # type: ignore
    item = client.get_item(cls._pydynox_table, keys)  # type: ignore
    if item is None:
        return None
    return cls.model_validate(item)  # type: ignore


def _save_method(self: T) -> None:
    """Save the model to DynamoDB.

    Validates with Pydantic before saving.
    """
    client = self.__class__._get_client()  # type: ignore
    item = self.model_dump()  # type: ignore
    client.put_item(self.__class__._pydynox_table, item)  # type: ignore


def _delete_method(self: T) -> None:
    """Delete the model from DynamoDB."""
    client = self.__class__._get_client()  # type: ignore
    key = self._get_key()  # type: ignore
    client.delete_item(self.__class__._pydynox_table, key)  # type: ignore


def _update_method(self: T, **kwargs: Any) -> None:
    """Update specific attributes on the model.

    Updates both the local instance and DynamoDB.

    Args:
        **kwargs: Attribute values to update.
    """
    # Validate new values with Pydantic
    current_data = self.model_dump()  # type: ignore
    current_data.update(kwargs)
    validated = self.__class__.model_validate(current_data)  # type: ignore

    # Update local instance
    for attr_name, value in kwargs.items():
        setattr(self, attr_name, getattr(validated, attr_name))

    # Update in DynamoDB
    client = self.__class__._get_client()  # type: ignore
    key = self._get_key()  # type: ignore
    client.update_item(self.__class__._pydynox_table, key, updates=kwargs)  # type: ignore


def _get_key_method(self: T) -> dict[str, Any]:
    """Get the key dict for this instance."""
    cls = self.__class__
    key = {cls._pydynox_hash_key: getattr(self, cls._pydynox_hash_key)}  # type: ignore
    if cls._pydynox_range_key:  # type: ignore
        key[cls._pydynox_range_key] = getattr(self, cls._pydynox_range_key)  # type: ignore
    return key


def from_pydantic(
    model_class: type[T],
    table: str,
    hash_key: str,
    range_key: str | None = None,
    region: str | None = None,
    endpoint_url: str | None = None,
) -> Any:
    """Create a DynamoDB-enabled model from a Pydantic model.

    This is an alternative to the @dynamodb_model decorator.

    Args:
        model_class: The Pydantic model class.
        table: DynamoDB table name.
        hash_key: Name of the hash key attribute.
        range_key: Name of the range key attribute (optional).
        region: AWS region (optional).
        endpoint_url: Custom endpoint URL for local testing (optional).

    Returns:
        The model class with DynamoDB methods added.

    Example:
        >>> class User(BaseModel):
        ...     pk: str
        ...     sk: str
        ...     name: str
        >>>
        >>> UserDB = from_pydantic(User, table="users", hash_key="pk", range_key="sk")
        >>> user = UserDB(pk="USER#1", sk="PROFILE", name="John")
        >>> user.save()
    """
    return dynamodb_model(
        table=table,
        hash_key=hash_key,
        range_key=range_key,
        region=region,
        endpoint_url=endpoint_url,
    )(model_class)



"""Condition classes for building DynamoDB filter expressions.

Most users don't need to import from here. Just use attribute operators:

    User.age > 18
    User.status == "active"
    (User.age > 18) & (User.status == "active")

For dynamic condition building, import And, Or, Not:

    from pydynox.conditions import And
    conditions = [User.status == "active", User.age > 18]
    final = And(*conditions)
"""

from __future__ import annotations

from typing import Union

from pydynox._internal._conditions import (
    ConditionAnd,
    ConditionBeginsWith,
    ConditionBetween,
    ConditionComparison,
    ConditionContains,
    ConditionExists,
    ConditionIn,
    ConditionNot,
    ConditionNotExists,
    ConditionOr,
)

# Type alias for any condition
Condition = Union[
    ConditionComparison,
    ConditionBetween,
    ConditionIn,
    ConditionExists,
    ConditionNotExists,
    ConditionBeginsWith,
    ConditionContains,
    ConditionAnd,
    ConditionOr,
    ConditionNot,
]

__all__ = ["Condition", "And", "Or", "Not"]


def And(*conditions: Condition) -> ConditionAnd:
    """Combine conditions with AND.

    Args:
        *conditions: Two or more conditions to combine.

    Returns:
        Combined condition.

    Example:
        from pydynox.conditions import And
        cond = And(User.age > 18, User.status == "active")
    """
    if len(conditions) < 2:
        raise ValueError("And requires at least 2 conditions")
    result = ConditionAnd(conditions[0], conditions[1])
    for c in conditions[2:]:
        result = ConditionAnd(result, c)
    return result


def Or(*conditions: Condition) -> ConditionOr:
    """Combine conditions with OR.

    Args:
        *conditions: Two or more conditions to combine.

    Returns:
        Combined condition.

    Example:
        from pydynox.conditions import Or
        cond = Or(User.status == "active", User.status == "pending")
    """
    if len(conditions) < 2:
        raise ValueError("Or requires at least 2 conditions")
    result = ConditionOr(conditions[0], conditions[1])
    for c in conditions[2:]:
        result = ConditionOr(result, c)
    return result


def Not(condition: Condition) -> ConditionNot:
    """Negate a condition.

    Args:
        condition: Condition to negate.

    Returns:
        Negated condition.

    Example:
        from pydynox.conditions import Not
        cond = Not(User.deleted.exists())
    """
    return ConditionNot(condition)



from uuid import uuid4

from pydynox import Model, ModelConfig
from pydynox.attributes import ExpiresIn, StringAttribute, TTLAttribute


class Session(Model):
    model_config = ModelConfig(table="sessions")

    pk = StringAttribute(hash_key=True)
    user_id = StringAttribute()
    expires_at = TTLAttribute()


def create_session(user_id: str) -> Session:
    """Create a session that expires in 24 hours."""
    session = Session(
        pk=f"SESSION#{uuid4()}",
        user_id=user_id,
        expires_at=ExpiresIn.hours(24),
    )
    session.save()
    return session


def validate_session(session_id: str) -> bool:
    """Check if session is valid."""
    session = Session.get(pk=session_id)
    if not session or session.is_expired:
        return False
    return True


def refresh_session(session_id: str) -> None:
    """Extend session by 24 hours."""
    session = Session.get(pk=session_id)
    if session and not session.is_expired:
        session.extend_ttl(ExpiresIn.hours(24))



"""Tests for rate limiting."""

import time

import pytest


def test_fixed_rate_creation():
    """Test creating a FixedRate limiter."""
    from pydynox.rate_limit import FixedRate

    limiter = FixedRate(rcu=50)
    assert limiter.rcu == 50
    assert limiter.wcu is None


def test_fixed_rate_with_wcu():
    """Test FixedRate with both RCU and WCU."""
    from pydynox.rate_limit import FixedRate

    limiter = FixedRate(rcu=50, wcu=25)
    assert limiter.rcu == 50
    assert limiter.wcu == 25


def test_fixed_rate_metrics():
    """Test FixedRate metrics tracking."""
    from pydynox.rate_limit import FixedRate

    limiter = FixedRate(rcu=100, wcu=50)

    # Initial metrics should be zero
    assert limiter.consumed_rcu == 0
    assert limiter.consumed_wcu == 0
    assert limiter.throttle_count == 0

    # Acquire some capacity
    limiter._acquire_rcu(10.0)
    limiter._acquire_wcu(5.0)

    assert limiter.consumed_rcu == 10.0
    assert limiter.consumed_wcu == 5.0


def test_adaptive_rate_creation():
    """Test creating an AdaptiveRate limiter."""
    from pydynox.rate_limit import AdaptiveRate

    limiter = AdaptiveRate(max_rcu=100)

    # Should start at 50% of max
    assert limiter.current_rcu == 50.0
    assert limiter.max_rcu == 100.0


def test_adaptive_rate_with_wcu():
    """Test AdaptiveRate with both RCU and WCU."""
    from pydynox.rate_limit import AdaptiveRate

    limiter = AdaptiveRate(max_rcu=100, max_wcu=50)

    assert limiter.current_rcu == 50.0
    assert limiter.current_wcu == 25.0
    assert limiter.max_rcu == 100.0
    assert limiter.max_wcu == 50.0


def test_adaptive_rate_throttle_reduces_rate():
    """Test that throttle events reduce the rate."""
    from pydynox.rate_limit import AdaptiveRate

    limiter = AdaptiveRate(max_rcu=100)

    # Initial rate is 50
    assert limiter.current_rcu == 50.0

    # After throttle, should be 40 (50 * 0.8)
    limiter._on_throttle()
    assert limiter.current_rcu == 40.0

    # After another throttle, should be 32 (40 * 0.8)
    limiter._on_throttle()
    assert limiter.current_rcu == 32.0

    # Throttle count should be 2
    assert limiter.throttle_count == 2


def test_adaptive_rate_min_rcu():
    """Test that rate doesn't go below min_rcu."""
    from pydynox.rate_limit import AdaptiveRate

    limiter = AdaptiveRate(max_rcu=100, min_rcu=10)

    # Throttle many times
    for _ in range(20):
        limiter._on_throttle()

    # Should not go below min
    assert limiter.current_rcu >= 10.0


def test_fixed_rate_rate_limiting():
    """Test that FixedRate actually limits the rate."""
    from pydynox.rate_limit import FixedRate

    # Very low rate to make test fast
    limiter = FixedRate(rcu=10)

    # Acquire all tokens
    start = time.time()
    limiter._acquire_rcu(10.0)
    first_acquire = time.time() - start

    # First acquire should be instant (tokens available)
    assert first_acquire < 0.1

    # Second acquire should wait for refill
    start = time.time()
    limiter._acquire_rcu(5.0)
    second_acquire = time.time() - start

    # Should have waited ~0.5 seconds for 5 tokens at 10/sec
    assert second_acquire >= 0.4


@pytest.mark.parametrize(
    "rcu,wcu,burst",
    [
        pytest.param(50, None, None, id="rcu_only"),
        pytest.param(50, 25, None, id="rcu_and_wcu"),
        pytest.param(50, 25, 100, id="with_burst"),
    ],
)
def test_fixed_rate_configurations(rcu, wcu, burst):
    """Test various FixedRate configurations."""
    from pydynox.rate_limit import FixedRate

    limiter = FixedRate(rcu=rcu, wcu=wcu, burst=burst)
    assert limiter.rcu == rcu
    assert limiter.wcu == wcu


@pytest.mark.parametrize(
    "max_rcu,max_wcu,min_rcu",
    [
        pytest.param(100, None, None, id="rcu_only"),
        pytest.param(100, 50, None, id="rcu_and_wcu"),
        pytest.param(100, 50, 10, id="with_min"),
    ],
)
def test_adaptive_rate_configurations(max_rcu, max_wcu, min_rcu):
    """Test various AdaptiveRate configurations."""
    from pydynox.rate_limit import AdaptiveRate

    limiter = AdaptiveRate(max_rcu=max_rcu, max_wcu=max_wcu, min_rcu=min_rcu)
    assert limiter.max_rcu == max_rcu
    assert limiter.max_wcu == max_wcu
    # Should start at 50% of max
    assert limiter.current_rcu == max_rcu * 0.5



"""Tests for condition classes."""

import pytest
from pydynox.attributes import (
    ListAttribute,
    MapAttribute,
    NumberAttribute,
    StringAttribute,
)
from pydynox.conditions import And, Not, Or


def make_attr(cls, name):
    """Create test attribute with name set."""
    attr = cls()
    attr.attr_name = name
    return attr


# Comparison operators


def test_eq():
    name = make_attr(StringAttribute, "name")
    cond = name == "John"

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 = :v0"
    assert names == {"name": "#n0"}
    assert values == {":v0": "John"}


def test_ne():
    status = make_attr(StringAttribute, "status")
    cond = status != "deleted"

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 <> :v0"


def test_gt():
    age = make_attr(NumberAttribute, "age")
    cond = age > 18

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 > :v0"
    assert values == {":v0": 18}


def test_ge():
    age = make_attr(NumberAttribute, "age")
    cond = age >= 21

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 >= :v0"


def test_lt():
    price = make_attr(NumberAttribute, "price")
    cond = price < 100

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 < :v0"


def test_le():
    price = make_attr(NumberAttribute, "price")
    cond = price <= 50

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 <= :v0"


# Function conditions


def test_exists():
    email = make_attr(StringAttribute, "email")
    cond = email.exists()

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "attribute_exists(#n0)"
    assert names == {"email": "#n0"}


def test_does_not_exist():
    deleted = make_attr(StringAttribute, "deleted_at")
    cond = deleted.does_not_exist()

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "attribute_not_exists(#n0)"


def test_begins_with():
    sk = make_attr(StringAttribute, "sk")
    cond = sk.begins_with("ORDER#")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "begins_with(#n0, :v0)"
    assert values == {":v0": "ORDER#"}


def test_contains():
    tags = make_attr(ListAttribute, "tags")
    cond = tags.contains("premium")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "contains(#n0, :v0)"


def test_between():
    age = make_attr(NumberAttribute, "age")
    cond = age.between(18, 65)

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 BETWEEN :v0 AND :v1"
    assert values == {":v0": 18, ":v1": 65}


def test_is_in():
    status = make_attr(StringAttribute, "status")
    cond = status.is_in("active", "pending", "review")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0 IN (:v0, :v1, :v2)"
    assert values == {":v0": "active", ":v1": "pending", ":v2": "review"}


# Combined conditions


def test_and_operator():
    age = make_attr(NumberAttribute, "age")
    status = make_attr(StringAttribute, "status")
    cond = (age > 18) & (status == "active")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "(#n0 > :v0 AND #n1 = :v1)"
    assert names == {"age": "#n0", "status": "#n1"}


def test_or_operator():
    status = make_attr(StringAttribute, "status")
    cond = (status == "active") | (status == "pending")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "(#n0 = :v0 OR #n0 = :v1)"


def test_not_operator():
    deleted = make_attr(StringAttribute, "deleted")
    cond = ~deleted.exists()

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "(NOT attribute_exists(#n0))"


def test_complex_combination():
    age = make_attr(NumberAttribute, "age")
    status = make_attr(StringAttribute, "status")
    deleted = make_attr(StringAttribute, "deleted")

    cond = ((age > 18) & (status == "active")) | ~deleted.exists()

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "((#n0 > :v0 AND #n1 = :v1) OR (NOT attribute_exists(#n2)))"


# Nested access


def test_map_access():
    address = make_attr(MapAttribute, "address")
    cond = address["city"] == "NYC"

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0.#n1 = :v0"
    assert names == {"address": "#n0", "city": "#n1"}


def test_list_access():
    tags = make_attr(ListAttribute, "tags")
    cond = tags[0] == "premium"

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0[0] = :v0"


def test_deep_nested():
    data = make_attr(MapAttribute, "data")
    cond = data["users"][0]["name"] == "John"

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "#n0.#n1[0].#n2 = :v0"


# And, Or, Not functions


def test_and_function():
    age = make_attr(NumberAttribute, "age")
    status = make_attr(StringAttribute, "status")
    active = make_attr(StringAttribute, "active")

    cond = And(age > 18, status == "active", active == True)  # noqa: E712

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "((#n0 > :v0 AND #n1 = :v1) AND #n2 = :v2)"


def test_or_function():
    status = make_attr(StringAttribute, "status")

    cond = Or(status == "a", status == "b", status == "c")

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "((#n0 = :v0 OR #n0 = :v1) OR #n0 = :v2)"


def test_not_function():
    deleted = make_attr(StringAttribute, "deleted")

    cond = Not(deleted.exists())

    names: dict = {}
    values: dict = {}
    result = cond.serialize(names, values)

    assert result == "(NOT attribute_exists(#n0))"


def test_and_requires_two_conditions():
    age = make_attr(NumberAttribute, "age")

    with pytest.raises(ValueError, match="at least 2"):
        And(age > 18)


def test_or_requires_two_conditions():
    age = make_attr(NumberAttribute, "age")

    with pytest.raises(ValueError, match="at least 2"):
        Or(age > 18)



"""Shared fixtures for benchmarks."""

import os
import signal
import socket
import subprocess
import time

import boto3
import pytest
from pydynox import DynamoDBClient
from pynamodb.attributes import NumberAttribute, UnicodeAttribute
from pynamodb.models import Model

MOTO_PORT = 5557
MOTO_ENDPOINT = f"http://127.0.0.1:{MOTO_PORT}"


def _is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("127.0.0.1", port)) == 0


@pytest.fixture(scope="session")
def moto_server():
    """Start moto server for benchmarks."""
    proc = subprocess.Popen(
        ["uv", "run", "moto_server", "-p", str(MOTO_PORT)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        preexec_fn=os.setsid,
    )

    max_wait = 10
    waited = 0
    while not _is_port_in_use(MOTO_PORT) and waited < max_wait:
        time.sleep(0.5)
        waited += 0.5

    if not _is_port_in_use(MOTO_PORT):
        proc.terminate()
        pytest.fail("Moto server failed to start")

    yield proc

    try:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        proc.wait()
    except (ProcessLookupError, OSError):
        pass  # Process already terminated


@pytest.fixture(scope="session")
def boto_client(moto_server):
    """Create a boto3 DynamoDB client."""
    return boto3.client(
        "dynamodb",
        region_name="us-east-1",
        endpoint_url=MOTO_ENDPOINT,
        aws_access_key_id="testing",
        aws_secret_access_key="testing",
    )


@pytest.fixture(scope="session")
def bench_table(boto_client):
    """Create a DynamoDB table for benchmarks."""
    try:
        boto_client.delete_table(TableName="bench_table")
        time.sleep(0.5)
    except boto_client.exceptions.ResourceNotFoundException:
        pass

    boto_client.create_table(
        TableName="bench_table",
        KeySchema=[
            {"AttributeName": "pk", "KeyType": "HASH"},
            {"AttributeName": "sk", "KeyType": "RANGE"},
        ],
        AttributeDefinitions=[
            {"AttributeName": "pk", "AttributeType": "S"},
            {"AttributeName": "sk", "AttributeType": "S"},
        ],
        BillingMode="PAY_PER_REQUEST",
    )
    time.sleep(0.5)
    return boto_client


@pytest.fixture(scope="session")
def pydynox_client(bench_table):
    """Create a pydynox DynamoDBClient."""
    return DynamoDBClient(
        region="us-east-1",
        endpoint_url=MOTO_ENDPOINT,
        access_key="testing",
        secret_key="testing",
    )


class BenchModel(Model):
    """PynamoDB model for benchmarks."""

    class Meta:
        table_name = "bench_table"
        region = "us-east-1"
        host = MOTO_ENDPOINT
        aws_access_key_id = "testing"
        aws_secret_access_key = "testing"

    pk = UnicodeAttribute(hash_key=True)
    sk = UnicodeAttribute(range_key=True)
    name = UnicodeAttribute(null=True)
    age = NumberAttribute(null=True)
    email = UnicodeAttribute(null=True)
    status = UnicodeAttribute(null=True)


@pytest.fixture(scope="session")
def pynamodb_model(bench_table):
    """Return the PynamoDB model class."""
    # Force PynamoDB to describe the table and cache metadata
    if not BenchModel.exists():
        raise RuntimeError("Table should exist")
    return BenchModel



"""Client for local development."""

from pydynox import DynamoDBClient

# Connect to DynamoDB Local
client = DynamoDBClient(endpoint_url="http://localhost:8000")

# Connect to LocalStack
client = DynamoDBClient(endpoint_url="http://localhost:4566")



from pydynox import BatchWriter, DynamoDBClient

client = DynamoDBClient()

# Batch write - items are sent in groups of 25
with BatchWriter(client, "users") as batch:
    for i in range(100):
        batch.put({"pk": f"USER#{i}", "name": f"User {i}"})

# Mix puts and deletes
with BatchWriter(client, "users") as batch:
    batch.put({"pk": "USER#1", "name": "John"})
    batch.put({"pk": "USER#2", "name": "Jane"})
    batch.delete({"pk": "USER#3"})



from pydynox import DynamoDBClient

client = DynamoDBClient()

# Query returns a result object with .metrics
result = client.query(
    "users",
    key_condition_expression="#pk = :pk",
    expression_attribute_names={"#pk": "pk"},
    expression_attribute_values={":pk": "ORG#123"},
)

# Iterate over results
for item in result:
    print(item["name"])

# Access metrics after iteration
print(result.metrics.duration_ms)  # 45.2
print(result.metrics.consumed_rcu)  # 2.5
print(result.metrics.items_count)  # 10



"""Safe delete - only delete if conditions are met."""

from pydynox import Model, ModelConfig
from pydynox.attributes import NumberAttribute, StringAttribute


class Order(Model):
    model_config = ModelConfig(table="orders")

    pk = StringAttribute(hash_key=True)
    status = StringAttribute()
    total = NumberAttribute()


# Only delete if order is in "draft" status
order = Order.get(pk="ORDER#123")
order.delete(condition=Order.status == "draft")

# Can't delete orders that are already processed



"""Shared fixtures for integration tests."""

import os
import signal
import socket
import subprocess
import time

import boto3
import pytest
from pydynox import DynamoDBClient

MOTO_PORT = 5556
MOTO_ENDPOINT = f"http://127.0.0.1:{MOTO_PORT}"


def _is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("127.0.0.1", port)) == 0


@pytest.fixture(scope="session")
def moto_server():
    """Start moto server for the test session."""
    import sys

    if os.name == "nt":
        # Windows: use CREATE_NEW_PROCESS_GROUP for proper cleanup
        proc = subprocess.Popen(
            [sys.executable, "-m", "moto.server", "-p", str(MOTO_PORT)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )
    else:
        proc = subprocess.Popen(
            [sys.executable, "-m", "moto.server", "-p", str(MOTO_PORT)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=os.setsid,
        )

    max_wait = 10
    waited = 0
    while not _is_port_in_use(MOTO_PORT) and waited < max_wait:
        time.sleep(0.5)
        waited += 0.5

    if not _is_port_in_use(MOTO_PORT):
        proc.terminate()
        pytest.fail("Moto server failed to start")

    yield proc

    if os.name == "nt":
        # Windows: kill the process tree
        subprocess.run(
            ["taskkill", "/F", "/T", "/PID", str(proc.pid)],
            capture_output=True,
        )
    else:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
    proc.wait(timeout=5)


@pytest.fixture
def boto_client(moto_server):
    """Create a boto3 DynamoDB client."""
    return boto3.client(
        "dynamodb",
        region_name="us-east-1",
        endpoint_url=MOTO_ENDPOINT,
        aws_access_key_id="testing",
        aws_secret_access_key="testing",
    )


@pytest.fixture
def table(boto_client):
    """Create a DynamoDB table for testing."""
    try:
        boto_client.delete_table(TableName="test_table")
        time.sleep(0.1)
    except boto_client.exceptions.ResourceNotFoundException:
        pass

    boto_client.create_table(
        TableName="test_table",
        KeySchema=[
            {"AttributeName": "pk", "KeyType": "HASH"},
            {"AttributeName": "sk", "KeyType": "RANGE"},
        ],
        AttributeDefinitions=[
            {"AttributeName": "pk", "AttributeType": "S"},
            {"AttributeName": "sk", "AttributeType": "S"},
        ],
        BillingMode="PAY_PER_REQUEST",
    )
    return boto_client


@pytest.fixture
def dynamo(table):
    """Create a pydynox DynamoDBClient."""
    return DynamoDBClient(
        region="us-east-1",
        endpoint_url=MOTO_ENDPOINT,
        access_key="testing",
        secret_key="testing",
    )



"""Integration tests for atomic update operations."""

import pytest
from pydynox import Model, ModelConfig, set_default_client
from pydynox.attributes import ListAttribute, NumberAttribute, StringAttribute
from pydynox.exceptions import ConditionCheckFailedError


class User(Model):
    model_config = ModelConfig(table="test_table")

    pk = StringAttribute(hash_key=True)
    sk = StringAttribute(range_key=True)
    name = StringAttribute()
    count = NumberAttribute()
    balance = NumberAttribute()
    tags = ListAttribute()


@pytest.fixture(autouse=True)
def setup_client(dynamo):
    set_default_client(dynamo)


def test_atomic_add_increments_value(dynamo):
    user = User(pk="USER#1", sk="PROFILE", name="John", count=0)
    user.save()

    user.update(atomic=[User.count.add(1)])

    result = User.get(pk="USER#1", sk="PROFILE")
    assert result.count == 1


def test_atomic_add_decrements_with_negative(dynamo):
    user = User(pk="USER#2", sk="PROFILE", name="John", balance=100)
    user.save()

    user.update(atomic=[User.balance.add(-25)])

    result = User.get(pk="USER#2", sk="PROFILE")
    assert result.balance == 75


def test_atomic_set_updates_value(dynamo):
    user = User(pk="USER#3", sk="PROFILE", name="John", count=0)
    user.save()

    user.update(atomic=[User.name.set("Jane")])

    result = User.get(pk="USER#3", sk="PROFILE")
    assert result.name == "Jane"


def test_atomic_remove_deletes_attribute(dynamo):
    user = User(pk="USER#4", sk="PROFILE", name="John", count=10)
    user.save()

    user.update(atomic=[User.count.remove()])

    result = User.get(pk="USER#4", sk="PROFILE")
    assert result.count is None


def test_atomic_append_adds_to_list(dynamo):
    user = User(pk="USER#5", sk="PROFILE", name="John", tags=["a", "b"])
    user.save()

    user.update(atomic=[User.tags.append(["c", "d"])])

    result = User.get(pk="USER#5", sk="PROFILE")
    assert result.tags == ["a", "b", "c", "d"]


def test_atomic_prepend_adds_to_front(dynamo):
    user = User(pk="USER#6", sk="PROFILE", name="John", tags=["b", "c"])
    user.save()

    user.update(atomic=[User.tags.prepend(["a"])])

    result = User.get(pk="USER#6", sk="PROFILE")
    assert result.tags == ["a", "b", "c"]


def test_atomic_if_not_exists_sets_when_missing(dynamo):
    user = User(pk="USER#7", sk="PROFILE", name="John")
    user.save()

    user.update(atomic=[User.count.if_not_exists(100)])

    result = User.get(pk="USER#7", sk="PROFILE")
    assert result.count == 100


def test_atomic_if_not_exists_keeps_existing(dynamo):
    user = User(pk="USER#8", sk="PROFILE", name="John", count=50)
    user.save()

    user.update(atomic=[User.count.if_not_exists(100)])

    result = User.get(pk="USER#8", sk="PROFILE")
    assert result.count == 50


def test_multiple_atomic_operations(dynamo):
    user = User(pk="USER#9", sk="PROFILE", name="John", count=0, tags=["a"])
    user.save()

    user.update(
        atomic=[
            User.count.add(5),
            User.tags.append(["b"]),
            User.name.set("Jane"),
        ]
    )

    result = User.get(pk="USER#9", sk="PROFILE")
    assert result.count == 5
    assert result.tags == ["a", "b"]
    assert result.name == "Jane"


def test_atomic_with_condition_succeeds(dynamo):
    user = User(pk="USER#COND1", sk="PROFILE", name="John", balance=100)
    user.save()

    user.update(
        atomic=[User.balance.add(-50)],
        condition=User.balance >= 50,
    )

    result = User.get(pk="USER#COND1", sk="PROFILE")
    assert result.balance == 50


def test_atomic_with_condition_fails(dynamo):
    user = User(pk="USER#COND2", sk="PROFILE", name="John", balance=30)
    user.save()

    with pytest.raises(ConditionCheckFailedError):
        user.update(
            atomic=[User.balance.add(-50)],
            condition=User.balance >= 50,
        )

    result = User.get(pk="USER#COND2", sk="PROFILE")
    assert result.balance == 30


def test_atomic_set_and_remove_combined(dynamo):
    user = User(pk="USER#12", sk="PROFILE", name="John", count=10, balance=100)
    user.save()

    user.update(
        atomic=[
            User.name.set("Jane"),
            User.count.remove(),
        ]
    )

    result = User.get(pk="USER#12", sk="PROFILE")
    assert result.name == "Jane"
    assert result.count is None
    assert result.balance == 100


def test_atomic_add_with_string_condition(dynamo):
    user = User(pk="USER#COND3", sk="PROFILE", name="John", balance=200)
    user.save()

    user.update(
        atomic=[User.balance.add(-50)],
        condition=User.name == "John",
    )

    result = User.get(pk="USER#COND3", sk="PROFILE")
    assert result.balance == 150


def test_atomic_add_with_string_condition_fails(dynamo):
    user = User(pk="USER#COND4", sk="PROFILE", name="John", balance=200)
    user.save()

    with pytest.raises(ConditionCheckFailedError):
        user.update(
            atomic=[User.balance.add(-50)],
            condition=User.name == "Jane",
        )

    result = User.get(pk="USER#COND4", sk="PROFILE")
    assert result.balance == 200


def test_atomic_multiple_ops_with_condition(dynamo):
    user = User(pk="USER#COND5", sk="PROFILE", name="John", count=5, balance=100)
    user.save()

    user.update(
        atomic=[
            User.count.add(1),
            User.balance.add(-10),
        ],
        condition=User.balance >= 10,
    )

    result = User.get(pk="USER#COND5", sk="PROFILE")
    assert result.count == 6
    assert result.balance == 90


def test_atomic_append_with_condition(dynamo):
    user = User(pk="USER#COND6", sk="PROFILE", name="John", tags=["a"], count=10)
    user.save()

    user.update(
        atomic=[User.tags.append(["b"])],
        condition=User.count > 5,
    )

    result = User.get(pk="USER#COND6", sk="PROFILE")
    assert result.tags == ["a", "b"]


def test_atomic_set_with_exists_condition(dynamo):
    user = User(pk="USER#COND7", sk="PROFILE", name="John", count=10)
    user.save()

    user.update(
        atomic=[User.name.set("Jane")],
        condition=User.count.exists(),
    )

    result = User.get(pk="USER#COND7", sk="PROFILE")
    assert result.name == "Jane"


def test_atomic_with_not_exists_condition(dynamo):
    user = User(pk="USER#COND8", sk="PROFILE", name="John")
    user.save()

    user.update(
        atomic=[User.count.if_not_exists(100)],
        condition=User.count.does_not_exist(),
    )

    result = User.get(pk="USER#COND8", sk="PROFILE")
    assert result.count == 100



"""Safe balance transfer with condition."""

from pydynox import Model, ModelConfig
from pydynox.attributes import NumberAttribute, StringAttribute
from pydynox.exceptions import ConditionCheckFailedError


class Account(Model):
    model_config = ModelConfig(table="accounts")

    pk = StringAttribute(hash_key=True)  # account_id
    sk = StringAttribute(range_key=True)  # "BALANCE"
    balance = NumberAttribute()


def withdraw(account: Account, amount: int) -> bool:
    """Withdraw money only if balance is sufficient."""
    try:
        account.update(
            atomic=[Account.balance.add(-amount)],
            condition=Account.balance >= amount,
        )
        return True
    except ConditionCheckFailedError:
        return False


# Usage
account = Account(pk="ACC#123", sk="BALANCE", balance=100)
account.save()

# This succeeds - balance goes from 100 to 50
success = withdraw(account, 50)
print(f"Withdrew 50: {success}")  # True

# This fails - balance is 50, can't withdraw 75
success = withdraw(account, 75)
print(f"Withdrew 75: {success}")  # False



"""Type stubs for pydynox_core (Rust module)."""

from __future__ import annotations

from typing import Any

# Metrics
class OperationMetrics:
    duration_ms: float
    consumed_rcu: float | None
    consumed_wcu: float | None
    request_id: str | None
    items_count: int | None
    scanned_count: int | None

    def __init__(self, duration_ms: float = 0.0) -> None: ...

# Client
class DynamoDBClient:
    def __init__(
        self,
        region: str | None = None,
        access_key: str | None = None,
        secret_key: str | None = None,
        session_token: str | None = None,
        profile: str | None = None,
        endpoint_url: str | None = None,
    ) -> None: ...
    def get_region(self) -> str: ...
    def ping(self) -> bool: ...
    def put_item(
        self,
        table: str,
        item: dict[str, Any],
        condition_expression: str | None = None,
        expression_attribute_names: dict[str, str] | None = None,
        expression_attribute_values: dict[str, Any] | None = None,
    ) -> OperationMetrics: ...
    def get_item(
        self,
        table: str,
        key: dict[str, Any],
    ) -> tuple[dict[str, Any] | None, OperationMetrics]: ...
    def delete_item(
        self,
        table: str,
        key: dict[str, Any],
        condition_expression: str | None = None,
        expression_attribute_names: dict[str, str] | None = None,
        expression_attribute_values: dict[str, Any] | None = None,
    ) -> OperationMetrics: ...
    def update_item(
        self,
        table: str,
        key: dict[str, Any],
        updates: dict[str, Any] | None = None,
        update_expression: str | None = None,
        condition_expression: str | None = None,
        expression_attribute_names: dict[str, str] | None = None,
        expression_attribute_values: dict[str, Any] | None = None,
    ) -> OperationMetrics: ...
    def query_page(
        self,
        table: str,
        key_condition_expression: str,
        filter_expression: str | None = None,
        expression_attribute_names: dict[str, str] | None = None,
        expression_attribute_values: dict[str, Any] | None = None,
        limit: int | None = None,
        exclusive_start_key: dict[str, Any] | None = None,
        scan_index_forward: bool | None = None,
        index_name: str | None = None,
    ) -> tuple[list[dict[str, Any]], dict[str, Any] | None, OperationMetrics]: ...
    def batch_write(
        self,
        table: str,
        put_items: list[dict[str, Any]],
        delete_keys: list[dict[str, Any]],
    ) -> None: ...
    def batch_get(
        self,
        table: str,
        keys: list[dict[str, Any]],
    ) -> list[dict[str, Any]]: ...
    def transact_write(self, operations: list[dict[str, Any]]) -> None: ...
    def create_table(
        self,
        table_name: str,
        hash_key: tuple[str, str],
        range_key: tuple[str, str] | None = None,
        billing_mode: str = "PAY_PER_REQUEST",
        read_capacity: int | None = None,
        write_capacity: int | None = None,
        table_class: str | None = None,
        encryption: str | None = None,
        kms_key_id: str | None = None,
        wait: bool = False,
    ) -> None: ...
    def table_exists(self, table_name: str) -> bool: ...
    def delete_table(self, table_name: str) -> None: ...
    def wait_for_table_active(
        self,
        table_name: str,
        timeout_seconds: int | None = None,
    ) -> None: ...

# Rate limiting
class FixedRate:
    def __init__(
        self,
        rcu: int | None = None,
        wcu: int | None = None,
        burst: int | None = None,
    ) -> None: ...
    def _acquire_rcu(self, rcu: float) -> None: ...
    def _acquire_wcu(self, wcu: float) -> None: ...
    def _on_throttle(self) -> None: ...

class AdaptiveRate:
    def __init__(
        self,
        max_rcu: int,
        max_wcu: int | None = None,
        min_rcu: int = 1,
        min_wcu: int = 1,
    ) -> None: ...
    def _acquire_rcu(self, rcu: float) -> None: ...
    def _acquire_wcu(self, wcu: float) -> None: ...
    def _on_throttle(self) -> None: ...

class RateLimitMetrics:
    rcu_acquired: float
    wcu_acquired: float
    throttle_count: int

# Tracing
def enable_sdk_debug() -> None: ...

# Serialization
def py_to_dynamo(value: Any) -> dict[str, Any]: ...
def dynamo_to_py(value: dict[str, Any]) -> Any: ...
def item_to_dynamo(item: dict[str, Any]) -> dict[str, Any]: ...
def item_from_dynamo(item: dict[str, Any]) -> dict[str, Any]: ...

# Exceptions
class PydynoxError(Exception): ...
class TableNotFoundError(PydynoxError): ...
class TableAlreadyExistsError(PydynoxError): ...
class ValidationError(PydynoxError): ...
class ConditionCheckFailedError(PydynoxError): ...
class TransactionCanceledError(PydynoxError): ...
class ThrottlingError(PydynoxError): ...
class AccessDeniedError(PydynoxError): ...
class CredentialsError(PydynoxError): ...
class SerializationError(PydynoxError): ...
class ConnectionError(PydynoxError): ...
class EncryptionError(PydynoxError): ...

# Compression
class CompressionAlgorithm:
    Zstd: CompressionAlgorithm
    Lz4: CompressionAlgorithm
    Gzip: CompressionAlgorithm

def compress(
    data: bytes,
    algorithm: CompressionAlgorithm | None = None,
    level: int | None = None,
) -> bytes: ...
def decompress(
    data: bytes,
    algorithm: CompressionAlgorithm | None = None,
) -> bytes: ...
def should_compress(
    data: bytes,
    algorithm: CompressionAlgorithm | None = None,
    threshold: float | None = None,
) -> bool: ...
def compress_string(
    value: str,
    algorithm: CompressionAlgorithm | None = None,
    level: int | None = None,
    min_size: int | None = None,
    threshold: float | None = None,
) -> str: ...
def decompress_string(value: str) -> str: ...

# Encryption
class KmsEncryptor:
    key_id: str

    def __init__(
        self,
        key_id: str,
        region: str | None = None,
        context: dict[str, str] | None = None,
    ) -> None: ...
    def encrypt(self, plaintext: str) -> str: ...
    def decrypt(self, ciphertext: str) -> str: ...
    @staticmethod
    def is_encrypted(value: str) -> bool: ...



"""Tests for Pydantic integration."""

from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel, Field
from pydynox.integrations.pydantic import dynamodb_model, from_pydantic


@dynamodb_model(table="users", hash_key="pk", range_key="sk")
class User(BaseModel):
    """Test model for users."""

    pk: str
    sk: str
    name: str
    age: int = 0


def test_decorator_adds_metadata():
    """Decorator adds pydynox metadata to the class."""
    assert User._pydynox_table == "users"
    assert User._pydynox_hash_key == "pk"
    assert User._pydynox_range_key == "sk"


def test_decorator_adds_methods():
    """Decorator adds CRUD methods to the class."""
    assert hasattr(User, "get")
    assert hasattr(User, "save")
    assert hasattr(User, "delete")
    assert hasattr(User, "update")
    assert hasattr(User, "_get_key")


def test_model_still_works_as_pydantic():
    """Model still works as a normal Pydantic model."""
    user = User(pk="USER#1", sk="PROFILE", name="John", age=30)

    assert user.pk == "USER#1"
    assert user.name == "John"
    assert user.age == 30

    # Pydantic methods still work
    data = user.model_dump()
    assert data == {"pk": "USER#1", "sk": "PROFILE", "name": "John", "age": 30}


def test_model_validates_with_pydantic():
    """Model validates data with Pydantic."""
    # age must be int
    with pytest.raises(Exception):  # Pydantic ValidationError
        User(pk="USER#1", sk="PROFILE", name="John", age="not a number")


def test_get_key_returns_hash_and_range():
    """_get_key returns both hash and range key."""
    user = User(pk="USER#1", sk="PROFILE", name="John")

    key = user._get_key()

    assert key == {"pk": "USER#1", "sk": "PROFILE"}


@patch("pydynox.integrations.pydantic.DynamoDBClient")
def test_get_fetches_from_dynamodb(mock_client_class):
    """get() fetches item from DynamoDB and returns Pydantic model."""
    mock_client = MagicMock()
    mock_client.get_item.return_value = {
        "pk": "USER#1",
        "sk": "PROFILE",
        "name": "John",
        "age": 30,
    }
    mock_client_class.return_value = mock_client
    User._pydynox_client = None

    user = User.get(pk="USER#1", sk="PROFILE")

    assert user is not None
    assert isinstance(user, User)
    assert user.pk == "USER#1"
    assert user.name == "John"
    mock_client.get_item.assert_called_once_with("users", {"pk": "USER#1", "sk": "PROFILE"})


@patch("pydynox.integrations.pydantic.DynamoDBClient")
def test_get_returns_none_when_not_found(mock_client_class):
    """get() returns None when item not found."""
    mock_client = MagicMock()
    mock_client.get_item.return_value = None
    mock_client_class.return_value = mock_client
    User._pydynox_client = None

    user = User.get(pk="USER#1", sk="PROFILE")

    assert user is None


@patch("pydynox.integrations.pydantic.DynamoDBClient")
def test_save_puts_to_dynamodb(mock_client_class):
    """save() puts item to DynamoDB."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    User._pydynox_client = None

    user = User(pk="USER#1", sk="PROFILE", name="John", age=30)
    user.save()

    mock_client.put_item.assert_called_once_with(
        "users", {"pk": "USER#1", "sk": "PROFILE", "name": "John", "age": 30}
    )


@patch("pydynox.integrations.pydantic.DynamoDBClient")
def test_delete_removes_from_dynamodb(mock_client_class):
    """delete() removes item from DynamoDB."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    User._pydynox_client = None

    user = User(pk="USER#1", sk="PROFILE", name="John")
    user.delete()

    mock_client.delete_item.assert_called_once_with("users", {"pk": "USER#1", "sk": "PROFILE"})


@patch("pydynox.integrations.pydantic.DynamoDBClient")
def test_update_updates_dynamodb(mock_client_class):
    """update() updates item in DynamoDB."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    User._pydynox_client = None

    user = User(pk="USER#1", sk="PROFILE", name="John", age=30)
    user.update(name="Jane", age=31)

    # Local instance updated
    assert user.name == "Jane"
    assert user.age == 31

    # DynamoDB updated
    mock_client.update_item.assert_called_once_with(
        "users", {"pk": "USER#1", "sk": "PROFILE"}, updates={"name": "Jane", "age": 31}
    )


def test_from_pydantic_creates_model():
    """from_pydantic() creates a DynamoDB-enabled model."""

    class Product(BaseModel):
        pk: str
        name: str
        price: float

    ProductDB = from_pydantic(Product, table="products", hash_key="pk")

    assert ProductDB._pydynox_table == "products"
    assert ProductDB._pydynox_hash_key == "pk"
    assert hasattr(ProductDB, "save")


def test_decorator_requires_basemodel():
    """Decorator raises error if class is not a BaseModel."""
    with pytest.raises(TypeError, match="must be a Pydantic BaseModel"):

        @dynamodb_model(table="test", hash_key="id")
        class NotAModel:
            id: str


def test_hash_key_only_model():
    """Model with only hash key (no range key) works."""

    @dynamodb_model(table="simple", hash_key="id")
    class SimpleModel(BaseModel):
        id: str
        data: str

    item = SimpleModel(id="123", data="test")
    key = item._get_key()

    assert key == {"id": "123"}


def test_pydantic_field_validation():
    """Pydantic Field validation still works."""

    @dynamodb_model(table="validated", hash_key="id")
    class ValidatedModel(BaseModel):
        id: str
        age: int = Field(ge=0, le=150)

    # Valid
    item = ValidatedModel(id="1", age=30)
    assert item.age == 30

    # Invalid - age too high
    with pytest.raises(Exception):
        ValidatedModel(id="1", age=200)



"""Integration tests for put_item operation."""

import pytest

PUT_ITEM_CASES = [
    pytest.param(
        {"pk": "USER#1", "sk": "PROFILE", "name": "John", "email": "john@test.com"},
        id="strings",
    ),
    pytest.param(
        {"pk": "USER#2", "sk": "PROFILE", "age": 30, "score": 95.5},
        id="numbers",
    ),
    pytest.param(
        {"pk": "USER#3", "sk": "SETTINGS", "active": True, "dark_mode": False},
        id="booleans",
    ),
    pytest.param(
        {"pk": "USER#4", "sk": "PROFILE", "tags": ["admin", "active"], "scores": [100, 95]},
        id="lists",
    ),
    pytest.param(
        {"pk": "USER#5", "sk": "PROFILE", "address": {"city": "Seattle", "zip": "98101"}},
        id="nested_map",
    ),
    pytest.param(
        {
            "pk": "USER#6",
            "sk": "FULL",
            "name": "Test",
            "age": 25,
            "active": True,
            "tags": ["a", "b"],
            "meta": {"key": "value"},
        },
        id="mixed_types",
    ),
]


@pytest.mark.parametrize("item", PUT_ITEM_CASES)
def test_put_item_saves_correctly(dynamo, item):
    """Test saving items with different data types."""
    dynamo.put_item("test_table", item)

    # Verify by getting the item back
    key = {"pk": item["pk"], "sk": item["sk"]}
    result = dynamo.get_item("test_table", key)

    assert result is not None
    for field, value in item.items():
        assert result[field] == value


def test_put_item_overwrites_existing(dynamo):
    """Test that put_item overwrites an existing item."""
    item1 = {"pk": "USER#100", "sk": "PROFILE", "name": "Original"}
    item2 = {"pk": "USER#100", "sk": "PROFILE", "name": "Updated", "new_field": "value"}

    dynamo.put_item("test_table", item1)
    dynamo.put_item("test_table", item2)

    result = dynamo.get_item("test_table", {"pk": "USER#100", "sk": "PROFILE"})

    assert result["name"] == "Updated"
    assert result["new_field"] == "value"



from pydynox import DynamoDBClient

client = DynamoDBClient()

# get_item returns a dict with .metrics
item = client.get_item("users", {"pk": "USER#1"})

if item:
    print(item["name"])  # Works like a normal dict
    print(item.metrics.duration_ms)  # 12.1
    print(item.metrics.consumed_rcu)  # 0.5



"""Tests for ModelConfig and default client."""

from unittest.mock import MagicMock

import pytest
from pydynox import (
    Model,
    ModelConfig,
    clear_default_client,
    get_default_client,
    set_default_client,
)
from pydynox.attributes import StringAttribute


@pytest.fixture(autouse=True)
def reset_default_client():
    """Reset default client before and after each test."""
    clear_default_client()
    yield
    clear_default_client()


def test_model_config_required_table():
    """ModelConfig requires table name."""
    config = ModelConfig(table="users")
    assert config.table == "users"


def test_model_config_defaults():
    """ModelConfig has sensible defaults."""
    config = ModelConfig(table="users")

    assert config.client is None
    assert config.skip_hooks is False
    assert config.max_size is None


def test_model_config_with_client():
    """ModelConfig accepts a client."""
    mock_client = MagicMock()
    config = ModelConfig(table="users", client=mock_client)

    assert config.client is mock_client


def test_model_config_with_options():
    """ModelConfig accepts all options."""
    mock_client = MagicMock()
    config = ModelConfig(
        table="users",
        client=mock_client,
        skip_hooks=True,
        max_size=400000,
    )

    assert config.table == "users"
    assert config.client is mock_client
    assert config.skip_hooks is True
    assert config.max_size == 400000


def test_set_default_client():
    """set_default_client sets the global client."""
    mock_client = MagicMock()

    set_default_client(mock_client)

    assert get_default_client() is mock_client


def test_get_default_client_returns_none_when_not_set():
    """get_default_client returns None when no client is set."""
    assert get_default_client() is None


def test_clear_default_client():
    """clear_default_client removes the global client."""
    mock_client = MagicMock()
    set_default_client(mock_client)

    clear_default_client()

    assert get_default_client() is None


def test_model_uses_config_client():
    """Model uses client from model_config."""
    mock_client = MagicMock()
    mock_client.get_item.return_value = {"pk": "USER#1", "name": "John"}

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client)
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None  # Reset cached client

    User.get(pk="USER#1")

    mock_client.get_item.assert_called_once_with("users", {"pk": "USER#1"})


def test_model_uses_default_client_when_no_config_client():
    """Model uses default client when model_config.client is None."""
    mock_client = MagicMock()
    mock_client.get_item.return_value = {"pk": "USER#1", "name": "John"}
    set_default_client(mock_client)

    class User(Model):
        model_config = ModelConfig(table="users")
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None

    User.get(pk="USER#1")

    mock_client.get_item.assert_called_once_with("users", {"pk": "USER#1"})


def test_model_config_client_takes_priority_over_default():
    """model_config.client takes priority over default client."""
    default_client = MagicMock()
    config_client = MagicMock()
    config_client.get_item.return_value = {"pk": "USER#1", "name": "John"}

    set_default_client(default_client)

    class User(Model):
        model_config = ModelConfig(table="users", client=config_client)
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None

    User.get(pk="USER#1")

    config_client.get_item.assert_called_once()
    default_client.get_item.assert_not_called()


def test_model_raises_error_when_no_client():
    """Model raises error when no client is configured."""

    class User(Model):
        model_config = ModelConfig(table="users")
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None

    with pytest.raises(ValueError, match="No client configured"):
        User.get(pk="USER#1")


def test_model_raises_error_when_no_model_config():
    """Model raises error when model_config is not defined."""

    class User(Model):
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None
    mock_client = MagicMock()
    set_default_client(mock_client)

    with pytest.raises(ValueError, match="must define model_config"):
        User.get(pk="USER#1")


def test_model_skip_hooks_from_config():
    """Model respects skip_hooks from model_config."""
    mock_client = MagicMock()

    class User(Model):
        model_config = ModelConfig(table="users", client=mock_client, skip_hooks=True)
        pk = StringAttribute(hash_key=True)
        name = StringAttribute()

    User._client_instance = None
    user = User(pk="USER#1", name="John")

    # skip_hooks should be True from config
    assert user._should_skip_hooks(None) is True
    # But can be overridden per-call
    assert user._should_skip_hooks(False) is False


def test_model_get_table_from_config():
    """Model gets table name from model_config."""
    mock_client = MagicMock()

    class User(Model):
        model_config = ModelConfig(table="my_users_table", client=mock_client)
        pk = StringAttribute(hash_key=True)

    assert User._get_table() == "my_users_table"
